@@ -1,11 +1,14 @@
-use std::io::Write;
+use std::fmt;
+use std::io;
 use std::vec::Vec;
 
-use ansi_term::Colour::{Fixed, Green, Red, Yellow};
-use ansi_term::Style;
+use nu_ansi_term::Color::{Fixed, Green, Red, Yellow};
+use nu_ansi_term::Style;
 
 use bytesize::ByteSize;
 
+use console::AnsiCodeIterator;
+
 use syntect::easy::HighlightLines;
 use syntect::highlighting::Color;
 use syntect::highlighting::Theme;
@@ -13,8 +16,7 @@ use syntect::parsing::SyntaxSet;
 
 use content_inspector::ContentType;
 
-use encoding::all::{UTF_16BE, UTF_16LE};
-use encoding::{DecoderTrap, Encoding};
+use encoding_rs::{UTF_16BE, UTF_16LE};
 
 use unicode_width::UnicodeWidthChar;
 
@@ -31,23 +33,38 @@ use crate::line_range::RangeCheckResult;
 use crate::preprocessor::{expand_tabs, replace_nonprintable};
 use crate::style::StyleComponent;
 use crate::terminal::{as_terminal_escaped, to_ansi_color};
+use crate::vscreen::AnsiStyle;
 use crate::wrapping::WrappingMode;
 
+pub enum OutputHandle<'a> {
+    IoWrite(&'a mut dyn io::Write),
+    FmtWrite(&'a mut dyn fmt::Write),
+}
+
+impl<'a> OutputHandle<'a> {
+    fn write_fmt(&mut self, args: fmt::Arguments<'_>) -> Result<()> {
+        match self {
+            Self::IoWrite(handle) => handle.write_fmt(args).map_err(Into::into),
+            Self::FmtWrite(handle) => handle.write_fmt(args).map_err(Into::into),
+        }
+    }
+}
+
 pub(crate) trait Printer {
     fn print_header(
         &mut self,
-        handle: &mut dyn Write,
+        handle: &mut OutputHandle,
         input: &OpenedInput,
         add_header_padding: bool,
     ) -> Result<()>;
-    fn print_footer(&mut self, handle: &mut dyn Write, input: &OpenedInput) -> Result<()>;
+    fn print_footer(&mut self, handle: &mut OutputHandle, input: &OpenedInput) -> Result<()>;
 
-    fn print_snip(&mut self, handle: &mut dyn Write) -> Result<()>;
+    fn print_snip(&mut self, handle: &mut OutputHandle) -> Result<()>;
 
     fn print_line(
         &mut self,
         out_of_range: bool,
-        handle: &mut dyn Write,
+        handle: &mut OutputHandle,
         line_number: usize,
         line_buffer: &[u8],
     ) -> Result<()>;
@@ -66,34 +83,50 @@ impl<'a> SimplePrinter<'a> {
 impl<'a> Printer for SimplePrinter<'a> {
     fn print_header(
         &mut self,
-        _handle: &mut dyn Write,
+        _handle: &mut OutputHandle,
         _input: &OpenedInput,
         _add_header_padding: bool,
     ) -> Result<()> {
         Ok(())
     }
 
-    fn print_footer(&mut self, _handle: &mut dyn Write, _input: &OpenedInput) -> Result<()> {
+    fn print_footer(&mut self, _handle: &mut OutputHandle, _input: &OpenedInput) -> Result<()> {
         Ok(())
     }
 
-    fn print_snip(&mut self, _handle: &mut dyn Write) -> Result<()> {
+    fn print_snip(&mut self, _handle: &mut OutputHandle) -> Result<()> {
         Ok(())
     }
 
     fn print_line(
         &mut self,
         out_of_range: bool,
-        handle: &mut dyn Write,
+        handle: &mut OutputHandle,
         _line_number: usize,
         line_buffer: &[u8],
     ) -> Result<()> {
         if !out_of_range {
             if self.config.show_nonprintable {
-                let line = replace_nonprintable(line_buffer, self.config.tab_width);
+                let line = replace_nonprintable(
+                    line_buffer,
+                    self.config.tab_width,
+                    self.config.nonprintable_notation,
+                );
                 write!(handle, "{}", line)?;
             } else {
-                handle.write_all(line_buffer)?
+                match handle {
+                    OutputHandle::IoWrite(handle) => handle.write_all(line_buffer)?,
+                    OutputHandle::FmtWrite(handle) => {
+                        write!(
+                            handle,
+                            "{}",
+                            std::str::from_utf8(line_buffer).map_err(|_| Error::Msg(
+                                "encountered invalid utf8 while printing to non-io buffer"
+                                    .to_string()
+                            ))?
+                        )?;
+                    }
+                }
             };
         }
         Ok(())
@@ -119,6 +152,7 @@ pub(crate) struct InteractivePrinter<'a> {
     config: &'a Config<'a>,
     decorations: Vec<Box<dyn Decoration>>,
     panel_width: usize,
+    ansi_style: AnsiStyle,
     content_type: Option<ContentType>,
     #[cfg(feature = "git")]
     pub line_changes: &'a Option<LineChanges>,
@@ -202,6 +236,7 @@ impl<'a> InteractivePrinter<'a> {
             config,
             decorations,
             content_type: input.reader.content_type,
+            ansi_style: AnsiStyle::new(),
             #[cfg(feature = "git")]
             line_changes,
             highlighter_from_set,
@@ -209,7 +244,11 @@ impl<'a> InteractivePrinter<'a> {
         })
     }
 
-    fn print_horizontal_line_term(&mut self, handle: &mut dyn Write, style: Style) -> Result<()> {
+    fn print_horizontal_line_term(
+        &mut self,
+        handle: &mut OutputHandle,
+        style: Style,
+    ) -> Result<()> {
         writeln!(
             handle,
             "{}",
@@ -218,7 +257,7 @@ impl<'a> InteractivePrinter<'a> {
         Ok(())
     }
 
-    fn print_horizontal_line(&mut self, handle: &mut dyn Write, grid_char: char) -> Result<()> {
+    fn print_horizontal_line(&mut self, handle: &mut OutputHandle, grid_char: char) -> Result<()> {
         if self.panel_width == 0 {
             self.print_horizontal_line_term(handle, self.colors.grid)?;
         } else {
@@ -248,7 +287,7 @@ impl<'a> InteractivePrinter<'a> {
         }
     }
 
-    fn print_header_component_indent(&mut self, handle: &mut dyn Write) -> std::io::Result<()> {
+    fn print_header_component_indent(&mut self, handle: &mut OutputHandle) -> Result<()> {
         if self.config.style_components.grid() {
             write!(
                 handle,
@@ -276,7 +315,7 @@ impl<'a> InteractivePrinter<'a> {
 impl<'a> Printer for InteractivePrinter<'a> {
     fn print_header(
         &mut self,
-        handle: &mut dyn Write,
+        handle: &mut OutputHandle,
         input: &OpenedInput,
         add_header_padding: bool,
     ) -> Result<()> {
@@ -375,7 +414,7 @@ impl<'a> Printer for InteractivePrinter<'a> {
         Ok(())
     }
 
-    fn print_footer(&mut self, handle: &mut dyn Write, _input: &OpenedInput) -> Result<()> {
+    fn print_footer(&mut self, handle: &mut OutputHandle, _input: &OpenedInput) -> Result<()> {
         if self.config.style_components.grid()
             && (self.content_type.map_or(false, |c| c.is_text()) || self.config.show_nonprintable)
         {
@@ -385,7 +424,7 @@ impl<'a> Printer for InteractivePrinter<'a> {
         }
     }
 
-    fn print_snip(&mut self, handle: &mut dyn Write) -> Result<()> {
+    fn print_snip(&mut self, handle: &mut OutputHandle) -> Result<()> {
         let panel = self.create_fake_panel(" ...");
         let panel_count = panel.chars().count();
 
@@ -412,33 +451,35 @@ impl<'a> Printer for InteractivePrinter<'a> {
     fn print_line(
         &mut self,
         out_of_range: bool,
-        handle: &mut dyn Write,
+        handle: &mut OutputHandle,
         line_number: usize,
         line_buffer: &[u8],
     ) -> Result<()> {
         let line = if self.config.show_nonprintable {
-            replace_nonprintable(line_buffer, self.config.tab_width)
+            replace_nonprintable(
+                line_buffer,
+                self.config.tab_width,
+                self.config.nonprintable_notation,
+            )
+            .into()
         } else {
-            let line = match self.content_type {
+            match self.content_type {
                 Some(ContentType::BINARY) | None => {
                     return Ok(());
                 }
-                Some(ContentType::UTF_16LE) => UTF_16LE
-                    .decode(line_buffer, DecoderTrap::Replace)
-                    .map_err(|_| "Invalid UTF-16LE")?,
-                Some(ContentType::UTF_16BE) => UTF_16BE
-                    .decode(line_buffer, DecoderTrap::Replace)
-                    .map_err(|_| "Invalid UTF-16BE")?,
-                _ => String::from_utf8_lossy(line_buffer).to_string(),
-            };
-            // Remove byte order mark from the first line if it exists
-            if line_number == 1 {
-                match line.strip_prefix('\u{feff}') {
-                    Some(stripped) => stripped.to_string(),
-                    None => line,
+                Some(ContentType::UTF_16LE) => UTF_16LE.decode_with_bom_removal(line_buffer).0,
+                Some(ContentType::UTF_16BE) => UTF_16BE.decode_with_bom_removal(line_buffer).0,
+                _ => {
+                    let line = String::from_utf8_lossy(line_buffer);
+                    if line_number == 1 {
+                        match line.strip_prefix('\u{feff}') {
+                            Some(stripped) => stripped.to_string().into(),
+                            None => line,
+                        }
+                    } else {
+                        line
+                    }
                 }
-            } else {
-                line
             }
         };
 
@@ -480,7 +521,7 @@ impl<'a> Printer for InteractivePrinter<'a> {
             self.config.highlighted_lines.0.check(line_number) == RangeCheckResult::InRange;
 
         if highlight_this_line && self.config.theme == "ansi" {
-            write!(handle, "\x1B[4m")?;
+            self.ansi_style.update("^[4m");
         }
 
         let background_color = self
@@ -507,37 +548,51 @@ impl<'a> Printer for InteractivePrinter<'a> {
             let italics = self.config.use_italic_text;
 
             for &(style, region) in &regions {
-                let text = &*self.preprocess(region, &mut cursor_total);
-                let text_trimmed = text.trim_end_matches(|c| c == '\r' || c == '\n');
+                let ansi_iterator = AnsiCodeIterator::new(region);
+                for chunk in ansi_iterator {
+                    match chunk {
+                        // ANSI escape passthrough.
+                        (ansi, true) => {
+                            self.ansi_style.update(ansi);
+                            write!(handle, "{}", ansi)?;
+                        }
 
-                write!(
-                    handle,
-                    "{}",
-                    as_terminal_escaped(
-                        style,
-                        text_trimmed,
-                        true_color,
-                        colored_output,
-                        italics,
-                        background_color
-                    )
-                )?;
+                        // Regular text.
+                        (text, false) => {
+                            let text = &*self.preprocess(text, &mut cursor_total);
+                            let text_trimmed = text.trim_end_matches(|c| c == '\r' || c == '\n');
 
-                if text.len() != text_trimmed.len() {
-                    if let Some(background_color) = background_color {
-                        let ansi_style = Style {
-                            background: to_ansi_color(background_color, true_color),
-                            ..Default::default()
-                        };
+                            write!(
+                                handle,
+                                "{}",
+                                as_terminal_escaped(
+                                    style,
+                                    &format!("{}{}", self.ansi_style, text_trimmed),
+                                    true_color,
+                                    colored_output,
+                                    italics,
+                                    background_color
+                                )
+                            )?;
 
-                        let width = if cursor_total <= cursor_max {
-                            cursor_max - cursor_total + 1
-                        } else {
-                            0
-                        };
-                        write!(handle, "{}", ansi_style.paint(" ".repeat(width)))?;
+                            if text.len() != text_trimmed.len() {
+                                if let Some(background_color) = background_color {
+                                    let ansi_style = Style {
+                                        background: to_ansi_color(background_color, true_color),
+                                        ..Default::default()
+                                    };
+
+                                    let width = if cursor_total <= cursor_max {
+                                        cursor_max - cursor_total + 1
+                                    } else {
+                                        0
+                                    };
+                                    write!(handle, "{}", ansi_style.paint(" ".repeat(width)))?;
+                                }
+                                write!(handle, "{}", &text[text_trimmed.len()..])?;
+                            }
+                        }
                     }
-                    write!(handle, "{}", &text[text_trimmed.len()..])?;
                 }
             }
 
@@ -546,82 +601,98 @@ impl<'a> Printer for InteractivePrinter<'a> {
             }
         } else {
             for &(style, region) in &regions {
-                let text = self.preprocess(
-                    region.trim_end_matches(|c| c == '\r' || c == '\n'),
-                    &mut cursor_total,
-                );
-
-                let mut max_width = cursor_max - cursor;
-
-                // line buffer (avoid calling write! for every character)
-                let mut line_buf = String::with_capacity(max_width * 4);
-
-                // Displayed width of line_buf
-                let mut current_width = 0;
-
-                for c in text.chars() {
-                    // calculate the displayed width for next character
-                    let cw = c.width().unwrap_or(0);
-                    current_width += cw;
-
-                    // if next character cannot be printed on this line,
-                    // flush the buffer.
-                    if current_width > max_width {
-                        // Generate wrap padding if not already generated.
-                        if panel_wrap.is_none() {
-                            panel_wrap = if self.panel_width > 0 {
-                                Some(format!(
-                                    "{} ",
-                                    self.decorations
-                                        .iter()
-                                        .map(|d| d.generate(line_number, true, self).text)
-                                        .collect::<Vec<String>>()
-                                        .join(" ")
-                                ))
-                            } else {
-                                Some("".to_string())
-                            }
+                let ansi_iterator = AnsiCodeIterator::new(region);
+                for chunk in ansi_iterator {
+                    match chunk {
+                        // ANSI escape passthrough.
+                        (ansi, true) => {
+                            self.ansi_style.update(ansi);
+                            write!(handle, "{}", ansi)?;
                         }
 
-                        // It wraps.
-                        write!(
-                            handle,
-                            "{}\n{}",
-                            as_terminal_escaped(
-                                style,
-                                &line_buf,
-                                self.config.true_color,
-                                self.config.colored_output,
-                                self.config.use_italic_text,
-                                background_color
-                            ),
-                            panel_wrap.clone().unwrap()
-                        )?;
+                        // Regular text.
+                        (text, false) => {
+                            let text = self.preprocess(
+                                text.trim_end_matches(|c| c == '\r' || c == '\n'),
+                                &mut cursor_total,
+                            );
 
-                        cursor = 0;
-                        max_width = cursor_max;
+                            let mut max_width = cursor_max - cursor;
 
-                        line_buf.clear();
-                        current_width = cw;
+                            // line buffer (avoid calling write! for every character)
+                            let mut line_buf = String::with_capacity(max_width * 4);
+
+                            // Displayed width of line_buf
+                            let mut current_width = 0;
+
+                            for c in text.chars() {
+                                // calculate the displayed width for next character
+                                let cw = c.width().unwrap_or(0);
+                                current_width += cw;
+
+                                // if next character cannot be printed on this line,
+                                // flush the buffer.
+                                if current_width > max_width {
+                                    // Generate wrap padding if not already generated.
+                                    if panel_wrap.is_none() {
+                                        panel_wrap = if self.panel_width > 0 {
+                                            Some(format!(
+                                                "{} ",
+                                                self.decorations
+                                                    .iter()
+                                                    .map(|d| d
+                                                        .generate(line_number, true, self)
+                                                        .text)
+                                                    .collect::<Vec<String>>()
+                                                    .join(" ")
+                                            ))
+                                        } else {
+                                            Some("".to_string())
+                                        }
+                                    }
+
+                                    // It wraps.
+                                    write!(
+                                        handle,
+                                        "{}\n{}",
+                                        as_terminal_escaped(
+                                            style,
+                                            &format!("{}{}", self.ansi_style, line_buf),
+                                            self.config.true_color,
+                                            self.config.colored_output,
+                                            self.config.use_italic_text,
+                                            background_color
+                                        ),
+                                        panel_wrap.clone().unwrap()
+                                    )?;
+
+                                    cursor = 0;
+                                    max_width = cursor_max;
+
+                                    line_buf.clear();
+                                    current_width = cw;
+                                }
+
+                                line_buf.push(c);
+                            }
+
+                            // flush the buffer
+                            cursor += current_width;
+                            write!(
+                                handle,
+                                "{}",
+                                as_terminal_escaped(
+                                    style,
+                                    &format!("{}{}", self.ansi_style, line_buf),
+                                    self.config.true_color,
+                                    self.config.colored_output,
+                                    self.config.use_italic_text,
+                                    background_color
+                                )
+                            )?;
+                        }
                     }
-
-                    line_buf.push(c);
                 }
-
-                // flush the buffer
-                cursor += current_width;
-                write!(
-                    handle,
-                    "{}",
-                    as_terminal_escaped(
-                        style,
-                        &line_buf,
-                        self.config.true_color,
-                        self.config.colored_output,
-                        self.config.use_italic_text,
-                        background_color
-                    )
-                )?;
             }
 
             if let Some(background_color) = background_color {
@@ -640,6 +711,7 @@ impl<'a> Printer for InteractivePrinter<'a> {
         }
 
         if highlight_this_line && self.config.theme == "ansi" {
+            self.ansi_style.update("^[24m");
             write!(handle, "\x1B[24m")?;
         }
 
